﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using SlimDX;

//Some namespace mappings
using Buffer = SlimDX.Direct3D11.Buffer;
using D2D    = SlimDX.Direct2D;
using GDI    = System.Drawing;
using DXGI   = SlimDX.DXGI;
using D3D10  = SlimDX.Direct3D10;
using D3D101 = SlimDX.Direct3D10_1;
using D3D11  = SlimDX.Direct3D11;

namespace Articles.DirectXDeviceHandling
{
    public static class GraphicsHelper
    {
        /// <summary>
        /// Creates a default SwapChain for the given target control.
        /// </summary>
        /// <param name="targetControl">Target control of the swap chain.</param>
        /// <param name="factory">Factory for SwapChain creation.</param>
        /// <param name="renderDevice">Graphics device.</param>
        public static DXGI.SwapChain CreateDefaultSwapChain(Control targetControl, DXGI.Factory factory, D3D11.Device renderDevice)
        {
            //Create description of the swap chain
            DXGI.SwapChainDescription swapChainDesc = new DXGI.SwapChainDescription();
            swapChainDesc.OutputHandle = targetControl.Handle;
            swapChainDesc.IsWindowed = true;
            swapChainDesc.BufferCount = 1;
            swapChainDesc.Flags = DXGI.SwapChainFlags.AllowModeSwitch;
            swapChainDesc.ModeDescription = new DXGI.ModeDescription(
                targetControl.Width,
                targetControl.Height,
                new Rational(60, 1),
                DXGI.Format.R8G8B8A8_UNorm);
            swapChainDesc.SampleDescription = new DXGI.SampleDescription(1, 0);
            swapChainDesc.SwapEffect = DXGI.SwapEffect.Discard;
            swapChainDesc.Usage = DXGI.Usage.RenderTargetOutput;

            //Create and return the swap chain and the render target
            return new DXGI.SwapChain(factory, renderDevice, swapChainDesc);
        }

        /// <summary>
        /// Creates a depth buffer texture with given width and height.
        /// </summary>
        /// <param name="device">Graphics device.</param>
        /// <param name="width">Width of generated texture.</param>
        /// <param name="height">Height of generated texture.</param>
        public static D3D11.Texture2D CreateDepthBufferTexture(D3D11.Device device, int width, int height)
        {
            D3D11.Texture2DDescription textureDescription = new D3D11.Texture2DDescription();
            textureDescription.Width = width;
            textureDescription.Height = height;
            textureDescription.MipLevels = 1;
            textureDescription.ArraySize = 1;
            textureDescription.Format = GraphicsCore.Current.TargetHardware == TargetHardware.DirectX11 ? DXGI.Format.D32_Float : DXGI.Format.D24_UNorm_S8_UInt;
            textureDescription.Usage = D3D11.ResourceUsage.Default;
            textureDescription.SampleDescription = new DXGI.SampleDescription(1, 0);
            textureDescription.BindFlags = D3D11.BindFlags.DepthStencil;
            textureDescription.CpuAccessFlags = D3D11.CpuAccessFlags.None;
            textureDescription.OptionFlags = D3D11.ResourceOptionFlags.None;

            return new D3D11.Texture2D(device, textureDescription);
        }

        /// <summary>
        /// Creates a default viewport for the given control.
        /// </summary>
        /// <param name="targetControl">Target control object.</param>
        public static D3D11.Viewport CreateDefaultViewport(Control targetControl)
        {
            return CreateDefaultViewport(
                targetControl.Width,
                targetControl.Height);
        }

        /// <summary>
        /// Creates a default viewport for the given width and height
        /// </summary>
        /// <param name="targetControl">Target control object.</param>
        public static D3D11.Viewport CreateDefaultViewport(int width, int height)
        {
            D3D11.Viewport result = new D3D11.Viewport(
                0f, 0f,
                (float)width, (float)height,
                0f, 1f);
            return result;
        }

        /// <summary>
        /// Disposes the given object and returns null.
        /// </summary>
        public static T DisposeGraphicsObject<T>(T objectToDispose)
            where T : class, IDisposable
        {
            if (objectToDispose == null) { return null; }

            try { objectToDispose.Dispose(); }
            catch (Exception) { }
            return null;
        }
    }
}
